﻿/*****************************************
 * Instructions
 *  - Replace 'uwuserid' with your uWaterloo User ID
 *  - Select the current calendar term and enter the year
 *  - List students with whom you had discussions and who helped you
 *
 * uWaterloo User ID:  csubrama @uwaterloo.ca
 * Submitted for ECE 250
 * Department of Electrical and Computer Engineering
 * University of Waterloo
 * Calender Term of Submission: Winter 2018
 *
 * By submitting this file, I affirm that
 * I am the author of all modifications to
 * the provided code.
 *
 * The following is a list of uWaterloo User IDs of those students
 * I had discussions with in preparing this project:
 *    -
 *
 * The following is a list of uWaterloo User IDs of those students
 * who helped me with this project (describe their help; e.g., debugging):
 *    -
 *****************************************/

#ifndef DOUBLE_HASH_TABLE_H
#define DOUBLE_HASH_TABLE_H

#ifndef nullptr
#define nullptr 0
#endif

#include "Exception.h"
#include "ece250.h"

enum bin_state_t { UNOCCUPIED, OCCUPIED, ERASED };

template <typename Type>
class Quadratic_hash_table {
	private:
		int count;
		int power;
		int array_size;
		int mask;
		Type *array;
		bin_state_t *occupied;

		//Counter to count number of occupied bins (with complemented erased bins) for finding load factor
		double lf_ctr;

		int hash( Type const & ) const;

	public:
		Quadratic_hash_table( int = 5 );
		~Quadratic_hash_table();
		int size() const;
		int capacity() const;
		double load_factor() const;
		bool empty() const;
		bool member( Type const & ) const;
		Type bin( int ) const;

		void print() const;

		void insert( Type const & );
		bool erase( Type const & );
		void clear();

	// Friends

	template <typename T>
	friend std::ostream &operator<<( std::ostream &, Quadratic_hash_table<T> const & );
};

// CONSTRUCTOR
// The constructor takes an argument m and creates a hash table with 2^m bins, indexed from 0 to 2m − 1. The
// default value of m is 5. Notice that you need to allocate and initialize two arrays, one that stores the values
// in the hash table, and the other that stores the status of the bins. 
template <typename Type>
Quadratic_hash_table<Type>::Quadratic_hash_table( int m ):
count( 0 ), power( m ),
array_size( 1 << power ),
mask( array_size - 1 ),
lf_ctr(0),
array( new Type[array_size] ),
occupied( new bin_state_t[array_size] ) {

	for ( int i = 0; i < array_size; ++i ) {
		occupied[i] = UNOCCUPIED;
	}
}

// DESTRUCTOR
// The destructor deletes the memory allocated for the hash table. Notice that the hash table has been
// represented using two arrays, and they both need to be deleted.
template <typename Type>
Quadratic_hash_table<Type>::~Quadratic_hash_table() {
	delete[] array;
	delete[] occupied;
}

//HASH FUNCTION
//This function returns a hash value (index of the array) to be determined for our object!
template <typename Type>
int Quadratic_hash_table<Type>::hash(Type const &obj) const {

	//Index (hash value) assigned as the remainder of array size with the object
	int index = static_cast<int>(obj) % array_size;

	//Return index (hash value). If less than zero, return (index + array_size)
	if (index < 0) {
		index = index + array_size;
		return index;
	}
	else {
		return index;
	}
}

// Return the number of elements currently stored in the hash table. 
template <typename Type>
int Quadratic_hash_table<Type>::size() const {
	return count;
}

// Return the number of bins in the hash table. 
template <typename Type>
int Quadratic_hash_table<Type>::capacity() const {
	return array_size;
}

// Return true if the hash table is empty, false otherwise. 
template <typename Type>
bool Quadratic_hash_table<Type>::empty() const {
	return (size() == 0);
}

// Returns ratio of occupied or erased bins to capacity
template <typename Type>
double Quadratic_hash_table<Type>::load_factor() const {
	return (double)(count+lf_ctr)/static_cast<double>(array_size);
}

// Return true if object obj is in the hash table and false otherwise. 
template <typename Type>
bool Quadratic_hash_table<Type>::member(Type const &obj) const {

	//Here we are using quadratic probing to see if the object exists!

	//Assigning a hash value to the object
	int index = hash(obj);

	for (int n = 0; n < array_size; n++) {

		//The object is 100% not in the table as the bin is unoccupied, hence return false!
		if (occupied[index] == UNOCCUPIED) {
			return false;
		}
		//Object found! Returns true!
		else if ((occupied[index] == OCCUPIED) && (array[index] == obj)) {
			return true;
		}
		//Continue probing...
		else {
			index = ((index + n) % array_size);
		}
	}

	//No matches found :( Returns false...
	return false;

}

// Returns the entry in bin n. 
template <typename Type>
Type Quadratic_hash_table<Type>::bin(int n) const {
	return array[n];
}

template <typename Type>
void Quadratic_hash_table<Type>::insert(Type const &obj){
	
	//Overflow, if the table is full!
	if (count == array_size) {
		throw overflow();
	}
	else {
		//does nothing if object is already in the table
		if (member(obj)) {

			}
		else {
			
			//get an index for the object to be inserted and start probing!
			int index = hash(obj);
			for (int i = 0; i < array_size; i++){
				//When we reach an unoccupied or erased bin, place the object in that bin and 
				//update variables. Break after this...
				if (occupied[index] != OCCUPIED){
					array[index] = obj;

					occupied[index] = OCCUPIED;
					count++;
					break;
				}
				//Else, continue to probe...
				else {
					index = ((index + i) % array_size);
				}
			}
		}
	}
}

template <typename Type>
bool Quadratic_hash_table<Type>::erase(Type const &obj) {

	//If the obj doesn't exist, return false!
	if (!member(obj)) {
		return false;
	}
	else {

		//If it exists, we use quadratic probing to find it!
		int index = hash(obj);
		for (int i = 0; i < array_size; i++) {
			//If the object is found, we initialize the bin as ERASED (lazy deletion) and update the variables!
			if (array[index] == obj) {
				occupied[index] = ERASED;
				count--;

				//Increment load factor counter...
				lf_ctr++;

				return true;
			}
			//Else, continue to probe...
			else {
				index = ((index + i) % array_size);
			}
		}
	}
}

// Remove all the elements in the hash table by assigning them as UNOCCUPIED. 
//Also updates variables to their initial values!
template <typename Type>
void Quadratic_hash_table<Type>::clear() {

	for (int i = 0; i < array_size; i++) {
		occupied[i] = UNOCCUPIED;
	}

	count = 0;
	lf_ctr = 0;
}

//TESTER FUNCTION
//Made by me to print the table contents during testing!
/*template <typename Type>
void Quadratic_hash_table<Type>::print() const {
	for (int i = 0; i < array_size; i++) {
		std::cout << "Value is: " << array[i] << std::endl;
		std::cout << "Status is: " << occupied[i] << std::endl;
	}
}*/

template <typename T>
std::ostream &operator<<( std::ostream &out, Quadratic_hash_table<T> const &hash ) {
	for ( int i = 0; i < hash.capacity(); ++i ) {
		if ( hash.occupied[i] == UNOCCUPIED ) {
			out << "- ";
		} else if ( hash.occupied[i] == ERASED ) {
			out << "x ";
		} else {
			out << hash.array[i] << ' ';
		}
	}

	return out;
}

#endif
